package com.simpledbbook.core;

import java.net.*;
import java.text.*;
import java.util.*;

import javax.crypto.Mac;

import com.simpledbbook.*;
import com.sun.org.apache.xerces.internal.impl.dv.util.Base64;

public class Request {
  private static final SimpleDateFormat TIMESTAMP;
  private final SortedMap<String, String> PARAMS;
  private final Mac HMAC;
  private final URL URL;

  static {
    String format = "yyyy-MM-dd'T'HH%3'A'mm%3'A'ss.SSS'Z'";
    TIMESTAMP = new SimpleDateFormat(format);
    TIMESTAMP.setTimeZone(TimeZone.getTimeZone("GMT"));
  }

  public Request(Mac hMac, String action, String key, URL url) {
    URL = url;
    HMAC = hMac;
    PARAMS = new TreeMap<String, String>();
    PARAMS.put("Version", "2009-04-15");
    PARAMS.put("Timestamp", TIMESTAMP.format(new Date()));
    PARAMS.put("SignatureVersion", "2");
    PARAMS.put("SignatureMethod", hMac.getAlgorithm());
    PARAMS.put("Action", action);
    PARAMS.put("AWSAccessKeyId", key);
  }

  public URL getURL() {
    return URL;
  }

  public void setDomain(String name) {
    PARAMS.put("DomainName", name);
  }

  public void setSelectExpression(String select) {
    PARAMS.put("SelectExpression", encode(select));
  }

  private String encode(String value) {
    String result = "";
    try {
      result = URLEncoder.encode(value, "UTF-8");
    } catch (java.io.UnsupportedEncodingException e) {
      e.printStackTrace(); // can't happen
    }
    result = result.replace("+", "%20");
    result = result.replace("*", "%2A");
    result = result.replace("%7E", "~");
    return result;
  }

  public void setNextToken(String nextToken) {
    PARAMS.put("NextToken", encode(nextToken));
  }

  public void setConsistentRead() {
    PARAMS.put("ConsistentRead", "true");
  }

  public void setItemName(String name) {
    PARAMS.put("ItemName", encode(name));
  }

  public void setExpectedAttribute(String name, String value) {
    PARAMS.put("Expected.1.Name", encode(name));
    PARAMS.put("Expected.1.Value", encode(value));
  }
  
  public void setExpectedToNotExist(String name) {
    PARAMS.put("Expected.1.Name", encode(name));
    PARAMS.put("Expected.1.Exists", "false");
  }

  public void addAttributeNames(String... atts) {
    for (int i = 0; i < atts.length; i++) {
      PARAMS.put("AttributeName." + i, encode(atts[i]));
    }
  }

  public void addAttributes(Attribute... atts) {
    addPrefixAttributes("", atts);
  }

  private void addPrefixAttributes(String prefix, Attribute... atts){
    for (int i = 0; i < atts.length; i++) {
      String finalPrefix = prefix + "Attribute." + i + ".";
      PARAMS.put(finalPrefix + "Value", encode(atts[i].getValue()));
      PARAMS.put(finalPrefix + "Name", encode(atts[i].getName()));
      if (atts[i].isReplace()) {
        PARAMS.put(finalPrefix + "Replace", "true");
      }
    }
  }

  public void addItems(List<Item> items) {
    for (int i = 0; i < items.size(); i++) {
      String prefix = "Item." + i + ".";
      PARAMS.put(prefix + "ItemName", encode(items.get(i).getName()));
      addPrefixAttributes(prefix, items.get(i).getAtts());
    }
  }

  void sign() {
    PARAMS.put("Signature", computeSignature());
  }

  private String computeSignature() {
    byte[] bytes = getStringToSign().getBytes();
    byte[] signature = null;
    synchronized (HMAC) {
      signature = HMAC.doFinal(bytes);
    }
    return encode(Base64.encode(signature));
  }

  private String getStringToSign() {
    StringBuilder toSign = new StringBuilder("POST\n");
    toSign.append(URL.getHost()).append("\n");
    toSign.append(URL.getPath()).append("\n");
    toSign.append(getParamString());
    return toSign.toString();
  }

  String getParamString() {
    StringBuilder result = new StringBuilder();
    for (Map.Entry<String, String> p : PARAMS.entrySet()) {
      result.append('&').append(p.getKey());
      result.append('=').append(p.getValue());
    }
    return result.substring(1); // omit the leading '&'
  }
}
