package com.simpledbbook.core;

import java.util.*;
import java.util.regex.*;

import com.simpledbbook.*;

public class Response {
  // single group expressions passed to matchSingles()
  private static final Pattern DOMAIN_REGEX = Pattern.compile(
    "<DomainName>(.*?)</DomainName>");

  private static final Pattern TOKEN_REGEX = Pattern.compile(
    "<NextToken>(.*?)</NextToken>");

  private static final Pattern ID_REGEX = Pattern.compile(
    "<RequestI[Dd]>(.*?)</RequestI[Dd]>");

  private static final Pattern USAGE_REGEX = Pattern.compile(
    "<BoxUsage>(.*?)</BoxUsage>");

  private static final Pattern ERROR_CODE_REGEX = Pattern.compile(
    "<Error><Code>(.*?)</Code>");

  private static final Pattern META_REGEX = Pattern.compile(
    "<DomainMetadataResult>(.*?)</DomainMetadataResult>");

  // double group expressions passed to matchDoubles()
  private static final Pattern ATTRIBUTE_REGEX = Pattern.compile(
    "<Attribute><Name>(.*?)</Name><Value>(.*?)</Value></Attribute>");

  private static final Pattern ITEM_REGEX = Pattern.compile(
    "<Item><Name>(.*?)</Name>(.*?)</Item>");

  private static final Pattern ANYTAG_REGEX = Pattern.compile(
    "<([A-Za-z0-9]+?)>([A-Za-z0-9.]*?)</");

  // XML response from SimpleDB
  private String body;

  // HTTP response code
  private int code;

  Response(String responseBody, int HttpCode) {
    body = responseBody;
    code = HttpCode;
  }
  
  public String getBody() {
    return body;
  }

  public int getCode() {
    return code;
  }

  public List<String> extractDomains() {
    return matchSingles(DOMAIN_REGEX);
  }

  // returns a list of all single group expression matches
  private List<String> matchSingles(Pattern pattern) {
    Matcher matcher = pattern.matcher(body);
    List<String> result = new ArrayList<String>();
    while (hasMoreMatches(matcher)) {
      result.add(matcher.group(1));
    }
    return result;
  }

  private boolean hasMoreMatches(Matcher matcher) {
    return matcher.find() && matcher.groupCount() > 0;
  }
  
  public String extractBoxUsage() {
    return extractResponseMetadata(USAGE_REGEX);
  }

  public String extractRequestId() {
    return extractResponseMetadata(ID_REGEX);
  }

  public String extractErrorCode() {
    return extractResponseMetadata(ERROR_CODE_REGEX);
  }

  private String extractResponseMetadata(Pattern regex) {
    List<String> result = matchSingles(regex);
    return result.size() > 0 ? result.get(0) : "";
  }

  public String extractNextToken() {
    List<String> token = matchSingles(TOKEN_REGEX);
    return token.size() > 0 ? token.get(0) : "";
  }

  // Select returns Items with nested Attributes
  public List<Item> extractItems() {
    List<Item> items = new ArrayList<Item>();
    for (Attribute row : matchDoubles(ITEM_REGEX, body)) {
      Item i = new Item(row.getName());
      i.addAttributes(matchDoubles(ATTRIBUTE_REGEX,row.getValue()));
      items.add(i);
    }
    return items;
  }

  // returns a list of double group expression matches
  // with group 1 in the name and group 2 in the value
  private List<Attribute> matchDoubles(Pattern pattern, String s) {
    Matcher matcher = pattern.matcher(s);
    List<Attribute> result = new ArrayList<Attribute>();
    while (hasMoreMatches(matcher)) {
      String attrName = matcher.group(1);
      String attrValue = decodeXMLEntities(matcher.group(2));
      result.add(new Attribute(attrName, attrValue));
    }
    return result;
  }

  public List<Attribute> extractAttributes() {
    return matchDoubles(ATTRIBUTE_REGEX, body);
  }

  private String decodeXMLEntities(String str) {
    String result = str.replace("&amp;", "&");
    result = result.replace("&lt;", "<");
    result = result.replace("&gt;", ">");
    return result;
  }

  // DomainMetadata handling returns all the children of
  // <DomainMetadataResult> in the form of a map
  public Map<String, Long> extractDomainMetadata() {
    List<String> singleMatch = matchSingles(META_REGEX);
    if (singleMatch.size() > 0) {
      String metaDataAttributes = singleMatch.get(0);
      return matchDoublesAsMap(ANYTAG_REGEX, metaDataAttributes);
    }
    return null;
  }

  private Map<String, Long> matchDoublesAsMap(Pattern p, String s) {
    Matcher matcher = p.matcher(s);
    Map<String, Long> result = new HashMap<String, Long>();
    while (hasMoreMatches(matcher)) {
      result.put(matcher.group(1), Long.parseLong(matcher.group(2)));
    }
    return result;
  }
}