package com.simpledbbook.tasklist;

import java.security.InvalidParameterException;
import java.text.SimpleDateFormat;
import java.util.*;

import com.simpledbbook.*;

import static com.simpledbbook.tasklist.TaskWorkspace.REVISION;

public class TaskService {
  private static final String TASK_SELECT = 
    "SELECT * FROM `%s` WHERE itemName() like \'%s:%s'";
  private static final SimpleDateFormat TIMESTAMP;
  private static final SimpleDateFormat DATESTAMP;
  private static final String USER_GROUP = "TASK_Users";
  private UserService userService;
  private SimpleDB sdb;
  private String domain;
  
  static {
    String timeFormat = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
    TIMESTAMP = new SimpleDateFormat(timeFormat);
    String dateFormat = "yyyy-MM-dd";
    DATESTAMP = new SimpleDateFormat(dateFormat);
    TIMESTAMP.setTimeZone(TimeZone.getTimeZone("GMT"));
  }

  public TaskService(SimpleDB sdb, String userDomain, 
      String appDomain) {
    this.sdb = sdb;
    this.domain = appDomain;
    this.userService = new UserService(sdb, userDomain);
  }

  public String registerNewUser(String name, String password) {
    if (! validateUserName(name)) {
      throw new InvalidParameterException(name);
    }
    if (! validatePassword(password)) {
      throw new InvalidParameterException(password);
    }
    userService.createUser(name, password, USER_GROUP);
    TaskWorkspace workspace = new TaskWorkspace(name);
    workspace.setSessionId(randomSessionId());
    sdb.put(domain, workspace.toItem());
    return workspace.getSessionId();
  }

  public static String now() {
    return TIMESTAMP.format(new Date());
  }
  
  public static String today() {
    return DATESTAMP.format(new Date());
  }

  private String randomSessionId() {
    return ((long)(Math.random() * 1000000000D + 1)) + "";
  }

  private boolean validateUserName(String name) {
    boolean valid = ensureLengthBetween(name, 5, 15);
    valid &= ensureAlphaNumeric(name);
    return valid;
  }

  private boolean validatePassword(String password) {
    return ensureLengthBetween(password, 5, 20);
  }

  private boolean ensureAlphaNumeric(String name) {
    for (char c: name.toCharArray()) {
      if (! Character.isLetterOrDigit(c)) {
        return false;
      }
    }
    return true;
  }

  private boolean ensureLengthBetween(String value, int min, int max){
    if (value.length() < min || value.length() > max) {
      return false;
    }
    return true;
  }

  public String addTask(String user, String date, String value, 
      String sessionId) {
    TaskWorkspace workspace = validateSession(user, sessionId);
    int taskId = incrementTaskId(workspace);
    int revision = incrementRevision(workspace);
    addTaskIntoWorkspace(date, workspace, taskId);
    sdb.putIfExists(domain, workspace.toItem(),REVISION,revision +"");
    storeTask(user, value, taskId);
    return taskId + "";
  }

  private int incrementRevision(TaskWorkspace workspace) {
    int revision = workspace.getWorkspaceRevision();
    workspace.setWorkspaceRevision(revision + 1);
    return revision;
  }

  private int incrementTaskId(TaskWorkspace workspace) {
    int taskId = workspace.getNextTaskId();
    workspace.setNextTaskId(taskId + 1);
    return taskId;
  }

  private void storeTask(String user, String value, int taskId) {
    Item newTask = new Item(user + ":" + taskId);
    newTask.addAttribute("text", value, true);
    sdb.put(domain, newTask); 
  }

  private void addTaskIntoWorkspace(String date,
      TaskWorkspace workspace, int taskId) {
    List<String> list = workspace.getDay(date);
    if (list == null) {
      list = new ArrayList<String>();
    }
    list.add(taskId + "");
    workspace.setDay(date, list);
  }
  
  private TaskWorkspace validateSession(String user, String sessionId){
    Item item = sdb.getConsistent(domain, user);
    String storedId = item.getAttribute("sessionId");
    if (storedId == null || ! sessionId.equals(storedId)) {
      throw new AuthenticationFailure();
    }
    return TaskWorkspace.fromItem(item);
  }

  public List<String> getTasksForDay(String user, String date, 
      String sessionId) {
    TaskWorkspace workspace = validateSession(user, sessionId);
    List<String> taskIds = workspace.getDay(date);
    String queryString = String.format(TASK_SELECT, domain, user, "%");
    List<Item> items = sdb.select(queryString);
    List<String> result = new ArrayList<String>();
    for (Item i: items) {
      String taskId = i.getName().substring(user.length() + 1);
      if (taskIds.contains(taskId)) {
        result.add(i.getAttribute("text"));
      }
    }
    return result;
  }
    
  public TaskWorkspace logIn(String user, String password) {
    List<String> groups = userService.authenticate(user, password);
    if (! groups.contains(USER_GROUP)) {
      throw new AuthenticationFailure();
    }
    Item item = sdb.get(domain, user);
    TaskWorkspace workspace = TaskWorkspace.fromItem(item);
    workspace.setSessionId(randomSessionId());
    sdb.put(domain, workspace.toItem());
    return workspace;
  }
  
  public void logOut(String user, String sessionId) {
    TaskWorkspace workspace = validateSession(user, sessionId);
    workspace.setSessionId("");
    sdb.put(domain, workspace.toItem());
  }
  
  public void deleteUser(String user) {
    userService.deleteUser(user);
    sdb.delete(domain, new Item(user));
  }
  
  public List<String> listUsers() {
    return userService.listUsers(USER_GROUP);
  }
}